@@ -28,6 +28,7 @@ var legacy_license_lib = require("./legacy_license_check.js");
 var {ResponseProcessor} = require("./bg/ResponseProcessor");
 var {Storage, ListStore} = require("./bg/Storage");
 var {ListManager} = require("./bg/ListManager");
+var {ExternalLicenses} = require("./bg/ExternalLicenses");
 
 console.log("main_background.js");
 /**
@@ -111,26 +112,6 @@ var reserved_objects = [
 	"eval"
 ];
 
-/**
-*	
-*	Sets global variable "webex" to either "chrome" or "browser" for
-*	use on Chrome or a Firefox variant.
-*
-*	Change this to support a new browser that isn't Chrome or Firefox,
-*	given that it supports webExtensions.
-*
-*	(Use the variable "webex" for all API calls after calling this)
-*/
-var webex;
-function set_webex(){
-	if(typeof(browser) == "object"){
-		webex = browser;
-	}
-	if(typeof(chrome) == "object"){
-		webex = chrome;
-	}
-}
-
 // Generates JSON key for local storage
 function get_storage_key(script_name,src_hash){
 	return script_name;
@@ -154,7 +135,7 @@ function options_listener(changes, area){
 	function flushed(){
 		dbg_print("cache flushed");
 	}	
-	//var flushingCache = webex.webRequest.handlerBehaviorChanged(flushed);
+	//var flushingCache = browser.webRequest.handlerBehaviorChanged(flushed);
 	
 
 	dbg_print("Items updated in area" + area +": ");
@@ -170,20 +151,23 @@ function options_listener(changes, area){
 }
 
 
-var active_connections = {};
-var unused_data = {};
-function createReport(initializer = null) {
+var activeMessagePorts = {};
+var activityReports = {};
+async function createReport(initializer = null) {
 	let template =  {
 		"accepted": [],
 		"blocked": [],
 		"blacklisted": [],
 		"whitelisted": [],
 		"unknown": [],
-		url: "",
 	};
 	if (initializer) {
 		template = Object.assign(template, initializer);
+		if (!template.url && initializer.tabId) {
+			template.url = (await browser.tabs.get(initializer.tabId)).url;
+		}
 	}
+	
 	template.site = ListStore.siteItem(template.url);
 	template.siteStatus = listManager.getStatus(template.site);
 	return template;
@@ -197,7 +181,7 @@ function createReport(initializer = null) {
 async function openReportInTab(data) {
 	let popupURL = await browser.browserAction.getPopup({});
 	let tab = await browser.tabs.create({url: `${popupURL}#fromTab=${data.tabId}`});
-	unused_data[tab.id] = createReport(data);
+	activityReports[tab.id] = await createReport(data);
 }
 
 /**
@@ -206,7 +190,7 @@ async function openReportInTab(data) {
 *
 */
 function debug_delete_local(){
-	webex.storage.local.clear();
+	browser.storage.local.clear();
 	dbg_print("Local storage cleared");
 }
 
@@ -222,9 +206,9 @@ function debug_print_local(){
 			console.log("%c "+i+" = "+items[i], 'color: blue;');
 		}
 	}
-	console.log("%c Variable 'unused_data': ", 'color: red;');
-	console.log(unused_data);
-	webex.storage.local.get(storage_got);
+	console.log("%c Variable 'activityReports': ", 'color: red;');
+	console.log(activityReports);
+	browser.storage.local.get(storage_got);
 }
 
 /**
@@ -242,9 +226,9 @@ function debug_print_local(){
 *	Make sure it will use the right URL when refering to a certain script.
 * 
 */
-function updateReport(tabId, oldReport, updateUI = false){
+async function updateReport(tabId, oldReport, updateUI = false){
 	let {url} = oldReport;
-	let newReport = createReport({url, tabId});
+	let newReport = await createReport({url, tabId});
 	for (let property of Object.keys(oldReport)) {
 		let entries = oldReport[property];
 		if (!Array.isArray(entries)) continue;
@@ -254,11 +238,12 @@ function updateReport(tabId, oldReport, updateUI = false){
 			if (Array.isArray(newReport[status])) newReport[status].push(script);
 		}
 	}
-	unused_data[tabId] = newReport;
+	activityReports[tabId] = newReport;
+	browser.sessions.setTabValue(tabId, url, newReport);
 	dbg_print(newReport);
-	if (updateUI && active_connections[tabId]) {
+	if (updateUI && activeMessagePorts[tabId]) {
 		dbg_print(`[TABID: ${tabId}] Sending script blocking report directly to browser action.`);
-		active_connections[tabId].postMessage({show_info: newReport});
+		activeMessagePorts[tabId].postMessage({show_info: newReport});
 	}
 }
 
@@ -282,10 +267,11 @@ function updateReport(tabId, oldReport, updateUI = false){
 *	Make sure it will use the right URL when refering to a certain script.
 *
 */
-async function addReportEntry(tabId, scriptHashOrUrl, action, update = false) {
-	if(!unused_data[tabId]) {
-		unused_data[tabId] = createReport({url: (await browser.tabs.get(tabId)).url});
-	}
+async function addReportEntry(tabId, scriptHashOrUrl, action) {
+	let report = activityReports[tabId];
+	if (!report) report = activityReports[tabId] = 
+			await createReport({tabId});
+			
 	let type, actionValue;
 	for (type of ["accepted", "blocked", "whitelisted", "blacklisted"]) {
 		if (type in action) {
@@ -310,10 +296,10 @@ async function addReportEntry(tabId, scriptHashOrUrl, action, update = false) {
 	let scriptName = actionValue[0];
 	try {
 		entryType = listManager.getStatus(scriptName, type);
-		let entries = unused_data[tabId][entryType];
+		let entries = report[entryType];
 		if(isNew(entries, scriptName)){
-			dbg_print(unused_data);
-			dbg_print(unused_data[tabId]);
+			dbg_print(activityReports);
+			dbg_print(activityReports[tabId]);
 			dbg_print(entryType);
 			entries.push(actionValue);
 		}
@@ -322,13 +308,15 @@ async function addReportEntry(tabId, scriptHashOrUrl, action, update = false) {
 		entryType = "unknown";
 	}
 	
-	if (active_connections[tabId]) {
+	if (activeMessagePorts[tabId]) {
 		try {
-			active_connections[tabId].postMessage({show_info: unused_data[tabId]});
+			activeMessagePorts[tabId].postMessage({show_info: report});
 		} catch(e) {
 		}
 	}
 	
+	browser.sessions.setTabValue(tabId, report.url, report);
+	updateBadge(tabId, report);
 	return entryType;
 }
 
@@ -358,7 +346,7 @@ function connected(p) {
 		function cb(items){
 			p.postMessage(items);
 		}
-		webex.storage.local.get(cb);
+		browser.storage.local.get(cb);
 		return;		
 	}
 	p.onMessage.addListener(async function(m) {
@@ -400,20 +388,20 @@ function connected(p) {
 			dbg_print(`[TABID:${tab.id}] Injecting contact finder`);
 			//inject_contact_finder(tabs[0]["id"]);
 		}
-		if (update || m.update && unused_data[m.tabId]) {
+		if (update || m.update && activityReports[m.tabId]) {
 			let tabId = "tabId" in m ?  m.tabId : tabs.pop().id;
 			dbg_print(`%c updating tab ${tabId}`, "color: red;");
-			active_connections[tabId] = p;
-			await updateReport(tabId, unused_data[tabId], true);
+			activeMessagePorts[tabId] = p;
+			await updateReport(tabId, activityReports[tabId], true);
 		} else {
 			for(let tab of tabs) {
-				if(unused_data[tab.id]){
+				if(activityReports[tab.id]){
 					// If we have some data stored here for this tabID, send it
 					dbg_print(`[TABID: ${tab.id}] Sending stored data associated with browser action'`);								
-					p.postMessage({"show_info": unused_data[tab.id]});
+					p.postMessage({"show_info": activityReports[tab.id]});
 				} else{
 					// create a new entry
-					let report = unused_data[tab.id] = createReport({"url": tab.url, tabId: tab.id});
+					let report = activityReports[tab.id] = await createReport({"url": tab.url, tabId: tab.id});
 					p.postMessage({show_info: report});							
 					dbg_print(`[TABID: ${tab.id}] No data found, creating a new entry for this window.`);	
 				}
@@ -430,12 +418,36 @@ function connected(p) {
 */
 function delete_removed_tab_info(tab_id, remove_info){
 	dbg_print("[TABID:"+tab_id+"]"+"Deleting stored info about closed tab");
-	if(unused_data[tab_id] !== undefined){
-		delete unused_data[tab_id];
+	if(activityReports[tab_id] !== undefined){
+		delete activityReports[tab_id];
 	}
-	if(active_connections[tab_id] !== undefined){
-		delete active_connections[tab_id];
+	if(activeMessagePorts[tab_id] !== undefined){
+		delete activeMessagePorts[tab_id];
 	}
+	ExternalLicenses.purgeCache(tab_id);
+}
+
+/**
+*	Called when the tab gets updated / activated
+*
+*	Here we check if  new tab's url matches activityReports[tabId].url, and if 
+* it doesn't we use the session cached value (if any).
+*
+*/
+
+async function onTabUpdated(tabId, changedInfo, tab) {
+	let {url} = tab;
+	let report = activityReports[tabId];
+	if (!(report && report.url === url)) {
+		let cache = await browser.sessions.getTabValue(tabId, url);
+		// on session restore tabIds may change
+		if (cache && cache.tabId !== tabId) cache.tabId = tabId; 
+		updateBadge(tabId, activityReports[tabId] = cache);
+	}
+}
+
+async function onTabActivated({tabId}) {
+	await onTabUpdated(tabId, {}, await browser.tabs.get(tabId));
 }
 
 /* *********************************************************************************************** */
@@ -741,7 +753,7 @@ function license_read(script_src, name, external = false){
 }
 
 /* *********************************************************************************************** */
-// TODO: Test if this script is being loaded from another domain compared to unused_data[tabid]["url"]
+// TODO: Test if this script is being loaded from another domain compared to activityReports[tabid]["url"]
 
 /**
 *	Asynchronous function, returns the final edited script as a string, 
@@ -774,19 +786,8 @@ async function get_script(response, url, tabId = -1, whitelisted = false, index
 	
 	let sourceHash = hash(response);
  	let domain = get_domain(url);
-	let report = unused_data[tabId] || (unused_data[tabId] = createReport({url, tabId}));
-	let blockedCount = report.blocked.length + report.blacklisted.length;
-	dbg_print(`amt. blocked on page: ${blockedCount}`);
-	if (blockedCount > 0 || !verdict) {
-		webex.browserAction.setBadgeText({
-			text: "!",
-			tabId
-		});
-		webex.browserAction.setBadgeBackgroundColor({
-			color: "red",
-			tabId
-		});
-	}
+	let report = activityReports[tabId] || (activityReports[tabId] = await createReport({tabId}));
+	updateBadge(tabId, report, !verdict);
 	let category = await addReportEntry(tabId, sourceHash, {"url": domain, [verdict ? "accepted" : "blocked"]: [url, reason]});
 	let scriptSource = verdict ? response : editedSource;
 	switch(category) {
@@ -798,6 +799,15 @@ async function get_script(response, url, tabId = -1, whitelisted = false, index
 	}
 }
 
+
+function updateBadge(tabId, report = null, forceRed = false) {
+	let blockedCount = report ? report.blocked.length + report.blacklisted.length : 0;
+	let [text, color] = blockedCount > 0 || forceRed 
+		? [blockedCount && blockedCount.toString() || "!" , "red"] : ["✓", "green"]
+	browser.browserAction.setBadgeText({text, tabId});
+	browser.browserAction.setBadgeBackgroundColor({color, tabId});
+}
+
 /**
 * 	Tests if a request is google analytics or not
 */
@@ -844,16 +854,16 @@ var ResponseHandler = {
 	/**
 	*	Enforce white/black lists for url/site early (hashes will be handled later)
 	*/
-	pre(response) {
+	async pre(response) {
 		let {request} = response;
-		let {url, type, tabId} = request;
+		let {url, type, tabId, frameId, documentUrl} = request;
 		
 		url = ListStore.urlItem(url);
 		let site = ListStore.siteItem(url);
 		
 		let blacklistedSite = blacklist.contains(site);
 		let blacklisted = blacklistedSite || blacklist.contains(url);
-		let topUrl = request.frameAncestors && request.frameAncestors.pop() || request.documentUrl;
+		let topUrl = request.frameAncestors && request.frameAncestors.pop() || documentUrl;
 		
 		if (blacklisted) {
 			if (type === "script") {
@@ -869,15 +879,31 @@ var ResponseHandler = {
 			});
 		} else {
 			let whitelistedSite = whitelist.contains(site);
-			if ((response.whitelisted = (whitelistedSite || whitelist.contains(url)))
-					&& type === "script") {
-				// accept the script and stop processing
-				addReportEntry(tabId, url, {url: topUrl, 
-					"whitelisted": [url, whitelistedSite ? `User whitelisted ${site}` : "Whitelisted by user"]});
-				return ResponseProcessor.ACCEPT;
+			let whitelisted = response.whitelisted = whitelistedSite || whitelist.contains(url);
+			if (type === "script") {
+				if (whitelisted) {
+					// accept the script and stop processing
+					addReportEntry(tabId, url, {url: topUrl, 
+						"whitelisted": [url, whitelistedSite ? `User whitelisted ${site}` : "Whitelisted by user"]});
+					return ResponseProcessor.ACCEPT;
+				} else {
+					let scriptInfo = await ExternalLicenses.check({url, tabId, frameId, documentUrl});
+					if (scriptInfo) {
+						let verdict, ret;
+						let msg = scriptInfo.toString();
+						if (scriptInfo.allFree) {
+							verdict = "accepted";
+							ret = ResponseProcessor.ACCEPT;
+						} else {
+							verdict = "blocked";
+							ret = ResponseProcessor.REJECT;
+						}
+						addReportEntry(tabId, url, {url, [verdict]: [url, msg]});
+						return ret;
+					}
+				}
 			}
 		}
-		
 		// it's a page (it's too early to report) or an unknown script:
 		//  let's keep processing
 		return ResponseProcessor.CONTINUE;
@@ -898,12 +924,28 @@ var ResponseHandler = {
 */
 async function handle_script(response, whitelisted){
 	let {text, request} = response;
-	let {url, tabId} = request;
+	let {url, tabId, frameId} = request;
 	url = ListStore.urlItem(url);
   let edited = await get_script(text, url, tabId, whitelisted, -2);
 	return Array.isArray(edited) ? edited[0] : edited;
 }
 
+/**
+* Serializes HTMLDocument objects including the root element and 
+*	the DOCTYPE declaration
+*/
+function doc2HTML(doc) {
+	let s = doc.documentElement.outerHTML;
+	if (doc.doctype) {
+		let dt = doc.doctype;
+		let sDoctype = `<!DOCTYPE ${dt.name || "html"}`;
+		if (dt.publicId) sDoctype += ` PUBLIC "${dt.publicId}"`;
+		if (dt.systemId) sDoctype += ` "${dt.systemId}"`;
+		s = `${sDoctype}>\n${s}`;
+	}
+	return s;
+}
+
 /**
 *	Removes noscript tags with name "librejs-path" leaving the inner content to load.
 */
@@ -914,7 +956,7 @@ function remove_noscripts(html_doc){
 		}
 	}
 	
-	return html_doc.documentElement.innerHTML;
+	return doc2HTML(html_doc);
 }
 
 /**
@@ -966,104 +1008,95 @@ function read_metadata(meta_element){
 			return false;
 		}
 }
-
 /**
+
 * 	Reads/changes the HTML of a page and the scripts within it.
 */
-function edit_html(html,url,tabid,wl){
+async function editHtml(html, documentUrl, tabId, frameId, whitelisted){
 	
-	return new Promise((resolve, reject) => {
-		if(wl == true){
-			// Don't bother, page is whitelisted
-			resolve(html);	 
-		}
+	var parser = new DOMParser();
+	var html_doc = parser.parseFromString(html, "text/html");
 		
-		var parser = new DOMParser();
-		var html_doc = parser.parseFromString(html, "text/html");
+	// moves external licenses reference, if any, before any <SCRIPT> element
+	ExternalLicenses.optimizeDocument(html_doc, {tabId, frameId, documentUrl}); 
+	
+	let url = ListStore.urlItem(documentUrl);
+	
+	if (whitelisted) { // don't bother rewriting
+		await get_script(html, url, tabId, whitelisted); // generates whitelisted report
+		return null;
+	}
 
-		var amt_scripts = 0;
-		var total_scripts = 0;
-		var scripts = html_doc.scripts;
+	var scripts = html_doc.scripts;
 		
-		var meta_element = html_doc.getElementById("LibreJS-info");
-		var first_script_src = "";
+	var meta_element = html_doc.getElementById("LibreJS-info");
+	var first_script_src = "";
 		
-		// get the potential inline source that can contain a license
-		for(var i = 0; i < scripts.length; i++){
-			// The script must be in-line and exist
-			if(scripts[i] !== undefined && scripts[i].src == ""){
-				first_script_src = scripts[i].innerHTML;
-				break;
-			}
+	// get the potential inline source that can contain a license
+	for (let script of scripts) {
+		// The script must be in-line and exist
+		if(script && !script.src){
+			first_script_src = script.textContent;
+			break;
 		}
+	}
 
-		var license = false;
-		if (first_script_src != "")
-			license = legacy_license_lib.check(first_script_src);
-		if(read_metadata(meta_element) || license != false ){
-			console.log("Valid license for intrinsic events found");
-			addReportEntry(tabid, url, {url, "accepted":[url, `Global license for the page: ${license}`]});
-			// Do not process inline scripts
-			scripts="";
-		}else{
-			// Deal with intrinsic events
-			var has_intrinsic_events = [];
-			for(var i = 0; i < html_doc.all.length; i++){
-				for(var j = 0; j < intrinsic_events.length; j++){
-					if(intrinsic_events[j] in html_doc.all[i].attributes){
-						has_intrinsic_events.push([i,j]);
+	let license = false;
+	if (first_script_src != "") {
+		license = legacy_license_lib.check(first_script_src);
+	}
+	if (read_metadata(meta_element) || license) {
+		console.log("Valid license for intrinsic events found");
+		addReportEntry(tabId, url, {url, "accepted":[url, `Global license for the page: ${license}`]});
+		// Do not process inline scripts
+		scripts = [];
+	} else {
+		let modified = false;
+		
+		// Deal with intrinsic events
+		for (let element of html_doc.all) {
+			let attributes = element.attributes;
+			for (let event of intrinsic_events) {
+				if (event in attributes) {
+					let attr = attributes[event];
+					try {
+						let edited = await get_script(attr.value, `Intrinsic event [${event}]`);
+					  if (edited) {
+							let value = edited[0];
+							if (value !== attr.value) {
+								modified = true;
+								attr.value = value;
+							}
+						}
+					} catch (e) {
+						console.error(e);
 					}
 				}
 			}
-
-			// "i" is an index in html_doc.all
-			// "j" is an index in intrinsic_events
-			function edit_event(src,i,j,name){
-				var edited = get_script(src, name);
-				edited.then(function(){
-					html_doc.all[i].attributes[intrinsic_events[j]].value = edited[0];
-				});
-			}
-
-			// Find all the document's elements with intrinsic events
-			for(var i = 0; i < has_intrinsic_events.length; i++){
-				var s_name = "Intrinsic event ["+has_intrinsic_events[i][0]+"]";
-				edit_event(html_doc.all[has_intrinsic_events[i][0]].attributes[intrinsic_events[has_intrinsic_events[i][1]]].value,has_intrinsic_events[i][0],has_intrinsic_events[i][1],s_name);
-			}
 		}
-
-		// Deal with inline scripts
-		for(var i = 0; i < scripts.length; i++){
-			if(scripts[i].src == ""){
-				total_scripts++;
-			}
-		}
-
-		dbg_print("Analyzing "+total_scripts+" inline scripts...");
-
-		for(var i = 0; i < scripts.length; i++){
-			if (scripts[i].src == ""){
-				if (scripts[i].type=="" || scripts[i].type=="text/javascript"){
-					var edit_script = get_script(scripts[i].innerHTML,url,tabid,wl,i);
-					edit_script.then(function(edited){
-						var edited_source = edited[0];
-						var unedited_source = html_doc.scripts[edited[1]].innerHTML.trim();
-						html_doc.scripts[edited[1]].innerHTML = edited_source;
-
-					});
-				}
-				amt_scripts++;
-				if(amt_scripts >= total_scripts){
-				resolve(remove_noscripts(html_doc));
+		
+		let modifiedInline = false;
+		for(let i = 0, len = scripts.length; i < len; i++) {
+			let script = scripts[i];
+			if (!script.src && !(script.type && script.type !== "text/javascript")) {
+				let edited = await get_script(script.textContent, url, tabId, whitelisted, i);
+				if (edited) {
+					let edited_source = edited[0];
+					let unedited_source = script.textContent.trim();
+					if (edited_source.trim() !== unedited_source) {
+						script.textContent = edited_source;
+						modified = modifiedInline = true;
+					}
 				}
 			}
+			if (modified) {
+				return modifiedInline 
+					? await remove_noscripts(html_doc) 
+					: doc2HTML(html_doc);
+			}
 		}
-		if(total_scripts == 0){
-			dbg_print("Nothing to analyze.");
-			resolve(remove_noscripts(html_doc));
-		}
-
-	});
+	}
+	return null;
 }
 
 /**
@@ -1071,20 +1104,12 @@ function edit_html(html,url,tabid,wl){
 */
 async function handle_html(response, whitelisted) {
 	let {text, request} = response;
-	let {url, tabId, type} = request;
-	url = ListStore.urlItem(url);
+	let {url, tabId, frameId, type} = request;
 	if (type === "main_frame") { 
-		delete unused_data[tabId];
-		browser.browserAction.setBadgeText({
-			text: "✓",
-			tabId
-		});
-		browser.browserAction.setBadgeBackgroundColor({
-			color: "green",
-			tabId
-		});
+		activityReports[tabId] = await createReport({url, tabId});
+		updateBadge(tabId);
 	}
-	return await edit_html(text, url, tabId, whitelisted);
+	return await editHtml(text, url, tabId, frameId, whitelisted);
 }
 
 var whitelist = new ListStore("pref_whitelist", Storage.CSV);
@@ -1102,11 +1127,11 @@ var listManager = new ListManager(whitelist, blacklist,
 */
 async function init_addon(){
 	await whitelist.load();
-	set_webex();
-	webex.runtime.onConnect.addListener(connected);
-	webex.storage.onChanged.addListener(options_listener);
-	webex.tabs.onRemoved.addListener(delete_removed_tab_info);
-
+	browser.runtime.onConnect.addListener(connected);
+	browser.storage.onChanged.addListener(options_listener);
+	browser.tabs.onRemoved.addListener(delete_removed_tab_info);
+	browser.tabs.onUpdated.addListener(onTabUpdated);
+	browser.tabs.onActivated.addListener(onTabActivated);
 	// Prevents Google Analytics from being loaded from Google servers
 	let all_types = [
 		"beacon", "csp_report", "font", "image", "imageset", "main_frame", "media",
@@ -1114,7 +1139,7 @@ async function init_addon(){
 		"web_manifest", "websocket", "xbl", "xml_dtd", "xmlhttprequest", "xslt", 
 		"other"
 	];
-	webex.webRequest.onBeforeRequest.addListener(
+	browser.webRequest.onBeforeRequest.addListener(
 		block_ga,
 		{urls: ["<all_urls>"], types: all_types},
 		["blocking"]
@@ -1134,7 +1159,7 @@ function inject_contact_finder(tab_id){
 	function executed(result) {
 	  dbg_print("[TABID:"+tab_id+"]"+"finished executing contact finder: " + result);
 	}
-	var executing = webex.tabs.executeScript(tab_id, {file: "/contact_finder.js"}, executed);
+	var executing = browser.tabs.executeScript(tab_id, {file: "/contact_finder.js"}, executed);
 }
 
 init_addon();
